/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.perseus.persistence.concurrency;

import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.perseus.concurrency.api.ConcurrencyException;
import org.objectweb.perseus.concurrency.distributed.DistributedConcurrencyManager;
import org.objectweb.perseus.concurrency.distributed.globallock.api.GlobalLockManager;
import org.objectweb.perseus.concurrency.distributed.globallock.lib.BasicGlobalLockManager;
import org.objectweb.perseus.concurrency.pessimistic.Lock;
import org.objectweb.perseus.concurrency.lib.RWLockValue;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.State;
import org.objectweb.perseus.persistence.api.StateManager;
import org.objectweb.perseus.persistence.api.StorageManager;
import org.objectweb.perseus.persistence.api.WorkingSet;
import org.objectweb.perseus.distribution.jgroups.JGroupsServiceManager;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class PDistributedConcurrencyManager
        extends DistributedConcurrencyManager {



    public final static String STORAGE_MANAGER_BINDING = "storage-manager";
    public final static String STATE_MANAGER_BINDING = "state-manager";

    protected StateManager stateManager;

    protected StorageManager storageManager;

    //IMPLEMENTATION OF THE UserBindingControler INTERFACE //
    //------------------------------------------------------//

    public String[] listFc() {
		String[] super_res = super.listFc();
		String res[] = new String[super_res.length + 2];
		for (int cpt = 0; cpt < super_res.length; cpt++) {
			res[cpt] = super_res[cpt];
		}
        res[super_res.length + 0] = STORAGE_MANAGER_BINDING;
		res[super_res.length + 1] = STATE_MANAGER_BINDING;
		return res;
    }

    public Object lookupFc(String s) throws NoSuchInterfaceException {
        if (STORAGE_MANAGER_BINDING.equals(s)) {
            return storageManager;
        } else if (STATE_MANAGER_BINDING.equals(s)) {
            return stateManager;
        } else {
            return super.lookupFc(s);
        }
    }

    public void bindFc(String s, Object o) throws IllegalBindingException,
			NoSuchInterfaceException {
        if (STORAGE_MANAGER_BINDING.equals(s)) {
            storageManager = (StorageManager) o;
        } else if (STATE_MANAGER_BINDING.equals(s)) {
            stateManager = (StateManager) o;
        } else super.bindFc(s, o);
    }

    public void unbindFc(String s) throws NoSuchInterfaceException {
        if (STORAGE_MANAGER_BINDING.equals(s)) {
            storageManager = null;
        } else if (STATE_MANAGER_BINDING.equals(s)) {
            stateManager = null;
        } else {
            super.unbindFc(s);
        }
    }


	public PDistributedConcurrencyManager() throws ConcurrencyException {
		super();
	}



	protected Object getState(Object ctx,
							  Object resourceId,
							  Lock lock,
							  Object hints) throws ConcurrencyException {
		CacheEntry ce = (CacheEntry) hints;
		WorkingSet ws = (WorkingSet) ctx;
        State state = stateManager.getReferenceState(ce);
        if (state == null) {
            synchronized (ce) {
                state = stateManager.getReferenceState(ce);
                if (state == null) {
                    state = stateManager.createState(ce);
                    stateManager.setReferenceState(state.getCacheEntry(), state);
                    try {
                        storageManager.read(ws, resourceId, state);
                        stateManager.makeClean(state);
                    } catch (PersistenceException e) {
                        stateManager.setReferenceState(state.getCacheEntry(), null);
                        stateManager.destroyState(state);
                        //The instance does not exist in the data support, remove the lock
                        synchronized (locks) {
                            if (lock.close(ctx)) {
                                locks.remove(lock);
                            }
                        }
                        throw new NoDSIConcurrencyException(e);
                    }
                }
            }
        }
        return state;
    }
	protected void invalidateState(Object resId, Object hints) {
		CacheEntry ce = (CacheEntry) hints;
		State state = stateManager.getReferenceState(ce);
		if (state != null) {
			synchronized(ce) {
				state = stateManager.getReferenceState(ce);
				if (state != null)
					stateManager.destroyState(state);
			}
		}
	}

}
